@@ -1,6 +1,6 @@
 /* $RCSfile: dag.c,v $
--- $Revision: 1.4 $
--- last change: $Author: rt $ $Date: 2004/09/08 16:05:07 $
+-- $Revision: 1.6.2.3 $
+-- last change: $Author: vq $ $Date: 2006/07/24 21:43:33 $
 --
 -- SYNOPSIS
 --      Routines to construct the internal dag.
@@ -29,7 +29,11 @@
 #include "extern.h"
 
 static void
-set_macro_value(hp)
+set_macro_value(hp)/*
+=====================
+  Set the macro according to its type. In addition to the string value
+  in hp->ht_value a macro can stores a value casted with its type.
+*/
 HASHPTR hp;
 {
    switch( hp->ht_flag & M_VAR_MASK )	/* only one var type per var */
@@ -48,8 +52,11 @@ HASHPTR hp;
          
          tvalue = atoi(hp->ht_value);
          if( hp->MV_IVAR == &Buffer_size ) {
-	    /* If Buffer_size is modified then make sure you change the
-             * size of the real buffer as well. */
+	    /* If Buffer_size (MAXLINELENGTH) is modified then make sure
+	     * you change the size of the real buffer as well. As the
+	     * value will at least be BUFSIZ this might lead to the
+	     * situation that the (string) value of MAXLINELENGTH is
+	     * smaller than the integer value. */
             tvalue = (tvalue < (BUFSIZ-2)) ? BUFSIZ : tvalue+2;
             if( Buffer_size == tvalue ) break;
             if( Buffer ) FREE(Buffer);
@@ -145,7 +152,13 @@ uint32  *phkey;
 
 
 PUBLIC HASHPTR
-Push_macro(hp)
+Push_macro(hp)/*
+================
+  This function pushes hp into the hash of all macros. If a previous
+  instance of the macro exists it is hidden by the new one. If one
+  existed before the new instance inherits some values from the preexisting
+  macro (see below for details).
+*/
 HASHPTR hp;
 {
    HASHPTR cur,prev;
@@ -154,20 +167,29 @@ HASHPTR hp;
 
    hv = Hash(hp->ht_name, &key);
 
+   /* Search for an existing instance of hp->ht_name, if found cur will point
+    * to it. */
    for(prev=NIL(HASH),cur=Macs[hv]; cur!=NIL(HASH); prev=cur,cur=cur->ht_next)
       if(    cur->ht_hash == key
 	  && !strcmp(cur->ht_name, hp->ht_name) )
          break;
 
    if (cur == NIL(HASH) || prev == NIL(HASH)) {
+      /* If no match or was found or the first element of Macs[hv] was
+       * the match insert hp at the beginning. */
       hp->ht_next  = Macs[hv];
       Macs[hv] = hp;
    }
    else {
+      /* otherwise insert hp in the chain. */
       hp->ht_next = prev->ht_next;
       prev->ht_next = hp;
    }
 
+   /* Inherit some parts of the former instance. Copying cur->var to hp->var
+    * copies the old value. Keeping the M_VAR_MASK (variable type) makes sure
+    * the type stays the same keeping M_PRECIOUS assures that the old values
+    * cannot be overridden if it is/was set. */
    if (cur) {
       memcpy((void *)&hp->var, (void *)&cur->var, sizeof(hp->var));
       hp->ht_flag |= ((M_VAR_MASK|M_PRECIOUS) & cur->ht_flag);
@@ -178,7 +200,12 @@ HASHPTR hp;
 
 
 PUBLIC HASHPTR
-Pop_macro(hp)
+Pop_macro(hp)/*
+================
+  This function pops (removes) hp from the hash of all macros. If a previous
+  instance of the macro existed it becomes accessible again.
+*/
+
 HASHPTR hp;
 {
    HASHPTR cur,prev;
@@ -187,23 +214,28 @@ HASHPTR hp;
 
    hv = Hash(hp->ht_name, &key);
 
+   /* Try to find hp. */ 
    for(prev=NIL(HASH),cur=Macs[hv]; cur != NIL(HASH);prev=cur,cur=cur->ht_next)
       if (cur == hp)
 	 break;
 
+   /* If cur == NIL macros was not found. */
    if (cur == NIL(HASH))
       return(NIL(HASH));
 
+   /* Remove hp from the linked list. */
    if (prev)
       prev->ht_next = cur->ht_next;
    else
       Macs[hv] = cur->ht_next;
 
+   /* Look for a previous (older) instance of hp->ht_name. */
    for(cur=cur->ht_next; cur != NIL(HASH); cur=cur->ht_next)
       if(    cur->ht_hash == key
 	  && !strcmp(cur->ht_name, hp->ht_name) )
          break;
 
+   /* If one was found we restore the typecast values. */
    if (cur)
       set_macro_value(cur);
 
@@ -278,9 +310,10 @@ int     flags;			/* initial ht_flags	*/
    if( (hp->ht_flag & M_USED) && !((flags | hp->ht_flag) & M_MULTI) )
       Warning( "Macro `%s' redefined after use", name );
 
+   /* If an empty string ("") is given set ht_value to NIL(char) */
    if( (value != NIL(char)) && (*value) ) {
       /* strip out any \<nl> combinations where \ is the current CONTINUATION
-       * char */
+       * char */ 
 
       for( p = value; (p = strchr(p, CONTINUATION_CHAR)) != NIL(char); )
          if( p[1] == '\n' )
@@ -316,6 +349,7 @@ int     flags;			/* initial ht_flags	*/
    /* Assign the hash table flag less the M_MULTI flag, it is used only
     * to silence the warning.  But carry it over if it was previously
     * defined in ht_flag, as this is a permanent M_MULTI variable. */
+   /* Also strip M_INIT flag if macro is assigned. */
 
    hp->ht_flag = (((flags & ~(M_MULTI|M_FORCE)) |
 		  (hp->ht_flag & (M_VAR_MASK | M_MULTI)))) & ~M_INIT;
@@ -519,12 +553,12 @@ int     fail;
 PUBLIC STRINGPTR
 Def_recipe( rcp, sp, white_too, no_check )/*
 =============================================
-	Take the recipe and add it to the list of recipes
-	pointed to by sp.  sp points to the last element.
-	return a pointer to the new recipe.  If white_too == TRUE add the
-	recipe even if it contains only white space.
-	If no_check is true then don't look for -@ at the start of the
-	recipe line. */
+   Take the recipe (rcp) and add it to the list of recipes pointed to by
+   sp (sp points to the last element).  If white_too == TRUE add the recipe
+   even if it contains only white space or an empty string.
+   Return a pointer to the new recipe (or sp if it was discarded).
+   If no_check is true then don't look for -@ at the start of the recipe
+   line. */
 char      *rcp;
 STRINGPTR sp;
 int       white_too;
