@@ -1,4 +1,4 @@
-/* RCS  $Id: ruletab.c,v 1.1.1.1 2000/09/22 15:33:30 hr Exp $
+/* RCS  $Id: ruletab.c,v 1.2 2006/04/20 12:08:38 hr Exp $
 --
 -- SYNOPSIS
 --      Default initial configuration of dmake.
@@ -36,7 +36,7 @@
 static char *_rules[] = {
 	"MAXLINELENGTH := 2046",
 	"MAXPROCESSLIMIT := 16",
-	".IMPORT .IGNORE: DMAKEROOT SOLARVER UPD INPATH OS UPDMINOREXT"
+	".IMPORT .IGNORE: DMAKEROOT"
 	".MAKEFILES : makefile.mk Makefile",
 	".SOURCE    : .NULL",
 #include "startup.h"
