@@ -1,6 +1,6 @@
 /* $RCSfile: ruletab.c,v $
--- $Revision: 1.4 $
--- last change: $Author: rt $ $Date: 2004/09/08 16:09:26 $
+-- $Revision: 1.6.2.1 $
+-- last change: $Author: vq $ $Date: 2006/08/14 20:35:56 $
 --
 -- SYNOPSIS
 --      Default initial configuration of dmake.
@@ -32,10 +32,10 @@
  * strictly so that dmake can parse the STARTUP makefile */
 
 static char *_rules[] = {
-	"MAXPROCESSLIMIT := 10",
-	"MAXLINELENGTH := 8190",
+	"MAXPROCESSLIMIT := 64",
+	"MAXLINELENGTH := 32766",
 #include "dmakeroot.h"
-	".IMPORT .IGNORE: DMAKEROOT SOLARVER UPD INPATH OS UPDMINOREXT",
+	".IMPORT .IGNORE: DMAKEROOT",
 	".MAKEFILES : makefile.mk Makefile makefile",
 	".SOURCE    : .NULL",
 #include "startup.h"
